"""
Python Wechaty - https://github.com/wechaty/python-wechaty
2020-now @ Copyright Wechaty

GitHub:
    TypeScript: https://github.com/wechaty/wechaty/blob/master/src/user/image.ts
    Python:     https://github.com/wechaty/python-wechaty/blob/master/src/wechaty/user/images.py

Authors:    Huan LI (李卓桓) <https://github.com/huan>
            Jingjing WU (吴京京) <https://github.com/wj-Mcat>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import annotations

from typing import (
    Type,
)


from wechaty_puppet import (    # type: ignore
    FileBox, ImageType, get_logger
)

from ..accessory import Accessory

log = get_logger('Image')


class Image(Accessory):
    """
    User Image class
    """

    def __str__(self):
        return 'Image<%d>' % self.image_id

    def __init__(
            self,
            image_id: str,
    ) -> None:
        """
        :param image_id:
        """
        super().__init__()
        log.info('init the message Image object <%s>', image_id)

        self.image_id = image_id

    @classmethod
    def create(cls: Type[Image], image_id: str) -> Image:
        """
        create image instance by image_id
        :param cls:
        :param image_id:
        :return:
        """
        log.info('@classmethod create(%s)', image_id)
        # obj = super().__new__(cls)
        # obj.__init__(image_id)
        # return obj
        return cls(image_id)

    async def thumbnail(self) -> FileBox:
        """
        docstring
        :return:
        """
        log.info('thumbnail() for <%s>', self.image_id)
        image_file = await self.puppet.message_image(
            message_id=self.image_id, image_type=ImageType.Thumbnail)
        return image_file

    async def hd(self) -> FileBox:
        """
        docstring
        :return:
        """
        log.info('hd() for <%s>', self.image_id)
        image_file = await self.puppet.message_image(
            message_id=self.image_id, image_type=ImageType.HD)
        return image_file

    async def artwork(self) -> FileBox:
        """
        docstring
        :return:
        """
        log.info('artwork() for <%d>', self.image_id)
        image_file = await self.puppet.message_image(
            message_id=self.image_id, image_type=ImageType.Artwork)
        return image_file



"""add type check utils"""
from typing import Union, Optional


def type_check(obj: object, type_name: str) -> bool:
    """
    circulation dependency problems can be resolved by TYPE_CHECKING,
    but this can not resolve NO type linting problems. eg:
        if isinstance(msg, Contact):
            pass
    in this problem, program don't import Contact at running time. So, it will
        throw a Exception, which will not be threw
    :param obj:
    :param type_name:
    :return:
    """
    if hasattr(obj, '__class__') and hasattr(obj.__class__, '__name__'):
        return obj.__class__.__name__ == type_name
    return False


def default_str(obj: Union[str, Optional[str]]) -> str:
    if obj:
        return obj
    return ''



"""
Tag for Contact Message
"""
from __future__ import annotations

from typing import (
    Dict,
    # Optional,
    Union,
    TYPE_CHECKING
)

from collections import defaultdict

from wechaty.exceptions import WechatyOperationError
from wechaty_puppet import get_logger  # type: ignore
from ..accessory import (
    Accessory,
)

if TYPE_CHECKING:
    from .contact import Contact
    from .favorite import Favorite

log = get_logger('Tag')


class Tag(Accessory):
    """
    tag object which handle the url_link content
    """
    _pool: Dict[str, 'Tag'] = defaultdict()

    tag_id: str

    def __init__(self, tag_id: str) -> None:
        """
        initialization for tag base class
        :param tag_id:
        """
        super().__init__()
        log.info('create tag %s', tag_id)

        self.tag_id = tag_id

    @classmethod
    def load(cls, tag_id: str) -> Tag:
        """
        load tag instance
        """
        if tag_id in cls._pool:
            return cls._pool[tag_id]

        new_tag = cls(tag_id)
        cls._pool[tag_id] = new_tag
        return new_tag

    @classmethod
    def get(cls, tag_id: str) -> Tag:
        """
        get tag objecr
        """
        log.info('load tag object %s', tag_id)
        return cls.load(tag_id)

    async def delete(self, target: Union[Contact, Favorite]):
        """
        remove tag from contact or favorite
        :param target:
        :return:
        """
        log.info('delete tag %s', self.tag_id)

        if target is Contact:
            await self.puppet.tag_contact_delete(tag_id=self.tag_id)
        elif target is Favorite:
            # TODO -> tag_favorite_delete not implement
            pass
            # await self.puppet.tag_contact_delete()
        else:
            raise WechatyOperationError('target param is required to be Contact or Favorite object')

    async def add(self, to: Union[Contact, Favorite]):
        """
        add tag to contact or favorite
        :param to:
        :return:
        """
        log.info('add tag to %s', str(to))
        if isinstance(to, Contact):
            await self.puppet.tag_contact_add(
                tag_id=self.tag_id, contact_id=to.contact_id
            )
        elif isinstance(to, Favorite):
            # TODO -> tag_favorite_add not implement
            pass
            # self.puppet.tag_favorite_add(self.tag_id, to)

    def remove(self, source: Union[Contact, Favorite]):
        """
        Remove this tag from Contact/Favorite

        tips : This function is depending on the Puppet Implementation,
        see [puppet-compatible-table](https://github.com/Chatie/
        wechaty/wiki/Puppet#3-puppet-compatible-table)
        :param source:
        :return:
        """
        log.info('remove tag for %s with %s',
                 self.tag_id,
                 str(source))
        try:
            if isinstance(source, Contact):
                self.puppet.tag_contact_remove(
                    tag_id=self.tag_id, contact_id=source.contact_id)
            elif isinstance(source, Favorite):
                # TODO -> tag_favorite_remove not implement
                pass
        except Exception as e:
            log.info('remove exception %s', str(e.args))
            raise WechatyOperationError('remove error')



"""doc"""

#
# import types from wechaty_puppet
#

from wechaty_puppet import (  # type: ignore
    FileBox,
    MessageType,
    MessagePayload,

    # Contact
    ContactGender,
    ContactType,
    ContactPayload,

    # Friendship
    FriendshipType,
    FriendshipPayload,

    # Room
    RoomPayload,
    RoomMemberPayload,

    # UrlLink

    # RoomInvitation
    RoomInvitationPayload,

    # Image
    ImageType,

    # Event
    EventType,

    RoomQueryFilter,
    RoomMemberQueryFilter,
    FriendshipSearchQueryFilter,
    ContactQueryFilter,
    MessageQueryFilter,
    ScanStatus
)

from .config import (
    get_logger,
)
from .accessory import Accessory
from .plugin import (
    WechatyPlugin,
    WechatyPluginOptions
)
from .wechaty import (
    Wechaty,
    WechatyOptions,
)
from .user import (
    Contact,
    Favorite,
    Friendship,
    Image,
    Message,
    MiniProgram,
    Room,
    RoomInvitation,
    Tag,
    UrlLink,
)
from .exceptions import (
    WechatyError,
    WechatyConfigurationError,
    WechatyAccessoryBindingError,
    WechatyStatusError,
    WechatyPayloadError,
    WechatyOperationError,
    WechatyPluginError,
)

from .version import VERSION

__version__ = VERSION

__all__ = [
    'Accessory',
    'Contact',
    'Favorite',
    'FileBox',
    'Friendship',
    'get_logger',
    'Image',
    'Message',
    'MiniProgram',
    'Room',
    'RoomInvitation',
    'Tag',
    'UrlLink',
    'Wechaty',
    'WechatyOptions',

    'WechatyPlugin',
    'WechatyPluginOptions',

    'MessageType',
    'MessagePayload',

    # Contact
    'ContactGender',
    'ContactType',
    'ContactPayload',

    # Friendship
    'FriendshipType',
    'FriendshipPayload',

    # Room
    'RoomPayload',
    'RoomMemberPayload',

    # UrlLink

    # RoomInvitation
    'RoomInvitationPayload',

    # Image
    'ImageType',

    # Event
    'EventType',

    'ScanStatus',

    'RoomQueryFilter',
    'RoomMemberQueryFilter',
    'FriendshipSearchQueryFilter',
    'ContactQueryFilter',
    'MessageQueryFilter',

    # Error
    'WechatyError',
    'WechatyConfigurationError',
    'WechatyAccessoryBindingError',
    'WechatyStatusError',
    'WechatyPayloadError',
    'WechatyOperationError',
    'WechatyPluginError',
]



"""
Python Wechaty - https://github.com/wechaty/python-wechaty

Authors:    Huan LI (李卓桓) <https://github.com/huan>
            Jingjing WU (吴京京) <https://github.com/wj-Mcat>

2020-now @ Copyright Wechaty

Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an 'AS IS' BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import annotations

from typing import (
    Union,
    Optional,
    TYPE_CHECKING
)
import json

from wechaty.exceptions import WechatyOperationError
from wechaty_puppet import (  # type: ignore
    FriendshipType,
    FriendshipPayload,
    get_logger
)
# from wechaty.utils import type_check

from ..types import Acceptable
from ..accessory import Accessory

if TYPE_CHECKING:
    from .contact import Contact

log = get_logger('FriendShip')


class Friendship(Accessory, Acceptable):
    """
    Send, receive friend request, and friend confirmation events.

    * 1. send request
    * 2. receive request(in friend event)
    * 3. confirmation friendship(friend event)
    """

    Type = FriendshipType

    def __init__(self, friendship_id: str):
        """
        initialization constructor for friendship
        """
        super(Friendship, self).__init__()

        self.friendship_id = friendship_id

        log.info('Friendship constructor %s', friendship_id)

    @classmethod
    def load(cls, friendship_id: str) -> Friendship:
        """
        load friendship without payload, which loads in a lazy way
        :param friendship_id:
        :return: initialized friendship
        """
        return cls(friendship_id)

    @classmethod
    async def search(cls, weixin: Optional[str] = None,
                     phone: Optional[str] = None) -> Optional[Contact]:
        """
        * Search a Friend by phone or weixin.
        *
        * The best practice is to search friend request once per minute.
        * Remeber not to do this too frequently, or your account
        * may be blocked.
        """
        log.info('search() <%s, %s>', cls, weixin, phone)
        friend_id = await cls.get_puppet().friendship_search(weixin=weixin,
                                                             phone=phone)
        if friend_id is None:
            return None
        contact = cls.get_wechaty().Contact.load(friend_id)
        await contact.ready()
        return contact

    @classmethod
    async def add(cls, contact: Contact, hello: str):
        """
        add friendship
        """
        log.info('add() <%s, %s>', contact.contact_id, hello)
        await cls.get_puppet().friendship_add(
            contact_id=contact.contact_id, hello=hello
        )

    @classmethod
    async def delete(cls, contact: Contact):
        """
        delete friendship
        """
        log.info('delete() a contact <%s>', contact.contact_id)
        log.warning('trying to delete a friend, which is dangerous, is not '
                    'implemented')
        # this is a dangerous action
        raise NotImplementedError

    def __str__(self) -> str:
        """
        string format for Friendship
        """
        if self._payload is None:
            return 'Friendship <{0}>'.format(self.friendship_id)
        return 'Friendship # type: {0}  contact: <{1}>  hello msg: <{2}>' \
            .format(self.type().name, self._payload.contact_id, self.hello())

    async def ready(self, force_sync: bool = False):
        """
        load friendship payload
        """
        log.info('ready() sync the friendship payload')
        if not self.is_ready() or force_sync:
            self._payload = await self.puppet.friendship_payload(
                friendship_id=self.friendship_id)

    def contact(self) -> Contact:
        """
        get the contact of the friendship
        """

        contact = self.wechaty.Contact.load(self.payload.contact_id)
        return contact

    async def accept(self):
        """
        accept friendship
        """
        log.info('accept friendship, friendship_id: <%s>', self.friendship_id)
        if self.type() != FriendshipType.FRIENDSHIP_TYPE_RECEIVE:
            raise WechatyOperationError(
                'accept() need type to be FriendshipType.'
                'FRIENDSHIP_TYPE_RECEIVE, but it got a " + FriendshipType : '
                '<{0}>'.format(self.type().name))

        log.info('friendship accept to %s', self.payload.contact_id)
        await self.puppet.friendship_accept(friendship_id=self.friendship_id)
        contact = self.contact()

        # reset contact data
        try:
            # TODO -> some other logical code
            await contact.ready()
        # pylint:disable=W0703
        except Exception as e:
            log.info(
                "can't reload contact data %s",
                str(e.args))

    def hello(self) -> str:
        """
        TODO ->
        Get verify message from
        """
        if self.payload.hello is None:
            hello_msg = ''
        else:
            hello_msg = self.payload.hello

        log.info('get hello message <%s> of friendship <%s>', hello_msg, self.payload)

        return hello_msg

    def type(self) -> FriendshipType:
        """
        Return the Friendship Type
        """
        if self.payload is None:
            return FriendshipType.FRIENDSHIP_TYPE_UNSPECIFIED
        if isinstance(self.payload.type, int):
            return FriendshipType(self.payload.type)
        elif isinstance(self.payload.type, FriendshipType):
            return self.payload.type
        else:
            raise TypeError('friendship type field type is limited between '
                            'int/FriendshipType')

    def to_json(self) -> str:
        """
        dumps the friendship
        """
        log.info('Friendship to_json')
        return json.dumps(self.payload)

    @classmethod
    async def from_json(
        cls,
        json_data: Union[str, FriendshipPayload]
    ) -> Friendship:
        """
        create friendShip by friendshipJson
        """
        log.info('from_json() <%s>', json_data)
        if isinstance(json_data, str):
            payload = FriendshipPayload(**json.loads(json_data))
        else:
            payload = json_data

        await cls.get_puppet().friendship_payload(
            friendship_id=payload.id, payload=payload
        )
        friendship = cls.get_wechaty().Friendship.load(payload.contact_id)
        await friendship.ready()
        return friendship



"""
Python Wechaty - https://github.com/wechaty/python-wechaty

Authors:    Huan LI (李卓桓) <https://github.com/huan>
            Jingjing WU (吴京京) <https://github.com/wj-Mcat>

2020-now @ Copyright Wechaty

Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an 'AS IS' BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import annotations

import dataclasses
import json
from typing import (
    Optional,
    Union,
    List,
)

from datetime import datetime

from wechaty.exceptions import WechatyPayloadError, WechatyOperationError
from wechaty_puppet import (  # type: ignore
    FileBox,
    MessagePayload,
    MessageQueryFilter,
    MessageType,
    get_logger
)

from ..accessory import Accessory
from .mini_program import MiniProgram
# TODO -> remove Sayable interface temporary
# from ..types import Sayable

from .contact import Contact
from .url_link import UrlLink
from .image import Image
from .room import Room

log = get_logger('Message')

SUPPORTED_MESSAGE_FILE_TYPES: List[MessageType] = [
    MessageType.MESSAGE_TYPE_ATTACHMENT,
    MessageType.MESSAGE_TYPE_EMOTICON,
    MessageType.MESSAGE_TYPE_IMAGE,
    MessageType.MESSAGE_TYPE_VIDEO
]


# pylint: disable=R0904,R0903
class Message(Accessory[MessagePayload]):
    """
    All wechat messages will be encapsulated as a Message.
    """

    Type = MessageType

    def __init__(self, message_id: str):
        """
        initialization
        """
        super().__init__()

        self.message_id = message_id

    def message_type(self) -> MessageType:
        """
        get message type
        """
        return self.payload.type

    def __str__(self) -> str:
        """
        format string for message
        """
        if not self.is_ready():
            return f'<{self.message_id}> not ready'

        room = self.room()
        room_format = '' if room is None \
            else f'room <{room}>, '

        to = self.to()
        to_format = '' if to is None \
            else f'say to contact <{to}>'
        return '%stalker %s, %s, %s' % (
            room_format,
            self.talker(),
            to_format,
            self.text(),
        )

    async def say(self, msg: Union[str, Contact, FileBox, UrlLink, MiniProgram],
                  mention_ids: Optional[List[str]] = None) -> Message:
        """
        send message
        """
        log.info('say() <%s>', msg)

        room = self.room()
        if room is not None:
            conversation_id = room.room_id
        else:
            talker = self.talker()
            if talker is None:
                raise WechatyPayloadError(f'Message must be from room/contact')
            conversation_id = talker.contact_id

        # in order to resolve circular dependency problems which is not for
        # typing, we import some modules locally.
        # TODO -> this is not good solution, we will fix it later.

        from .url_link import UrlLink
        from .mini_program import MiniProgram

        if isinstance(msg, str):
            message_id = await self.puppet.message_send_text(
                conversation_id=conversation_id,
                message=msg,
                mention_ids=mention_ids)

        elif isinstance(msg, Contact):
            message_id = await self.puppet.message_send_contact(
                conversation_id=conversation_id, contact_id=msg.contact_id)
        elif isinstance(msg, FileBox):
            message_id = await self.puppet.message_send_file(
                conversation_id=conversation_id, file=msg)
        elif isinstance(msg, UrlLink):
            message_id = await self.puppet.message_send_url(
                conversation_id=conversation_id, url=json.dumps(dataclasses.asdict(msg.payload)))
        elif isinstance(msg, MiniProgram):
            assert msg.payload is not None
            message_id = await self.puppet.message_send_mini_program(
                conversation_id=conversation_id,
                mini_program=msg.payload)
        else:
            raise WechatyPayloadError(f'message type should be str, '
                                      f'Contact/FileBox/UrlLink/MiniProgram')

        message = self.load(message_id)
        await message.ready()
        return message

    @classmethod
    async def find(cls, talker_id: Optional[str] = None,
                   message_id: Optional[str] = None,
                   room_id: Optional[str] = None,
                   text: Optional[str] = None,
                   to_id: Optional[str] = None,
                   message_type: Optional[MessageType] = None
                   ) -> Optional[Message]:
        """
        Find message in cache
        """
        log.info('Message find all <%s, %s, %s, <%s, %s, %s>', talker_id,
                 message_id, room_id, text, to_id, message_type)

        messages = await cls.find_all(
            talker_id=talker_id,
            message_id=message_id,
            room_id=room_id,
            text=text,
            to_id=to_id,
            message_type=message_type
        )
        if messages is None or len(messages) < 1:
            return None

        if len(messages) > 1:
            log.warning(
                'Message findAll() got more than one(%d) result',
                len(messages))
        return messages[0]

    @classmethod
    async def find_all(cls, talker_id: Optional[str] = None,
                       message_id: Optional[str] = None,
                       room_id: Optional[str] = None,
                       text: Optional[str] = None,
                       to_id: Optional[str] = None,
                       message_type: Optional[MessageType] = None
                       ) -> List[Message]:
        """
        Find messages in cache
        """
        log.info('Message find all <%s, %s, %s, <%s, %s, %s>', talker_id,
                 message_id, room_id, text, to_id, message_type)

        query_filter = MessageQueryFilter(
            from_id=talker_id,
            id=message_id,
            room_id=room_id,
            text=text,
            to_id=to_id,
            type=message_type
        )
        message_ids = await cls.get_puppet().message_search(query_filter)
        messages = [cls.load(message_id) for message_id in message_ids]
        return messages

    def talker(self) -> Contact:
        """
        get message talker

        # TODO ->   Suggestion: talker/to/room/text func can
        #           be converted to property func
        """
        talker_id = self.payload.from_id
        if talker_id is None:
            raise WechatyPayloadError(f'message must be from Contact')
        return self.wechaty.Contact.load(talker_id)

    def to(self) -> Optional[Contact]:
        """
        get message reply to
        """
        to_id = self.payload.to_id
        if to_id is None:
            return None
        return self.wechaty.Contact.load(to_id)

    def room(self) -> Optional[Room]:
        """
        get message room
        """
        room_id = self.payload.room_id
        if room_id is None or room_id == '':
            return None
        return self.wechaty.Room.load(room_id)

    def text(self) -> str:
        """
        get message text
        """
        if self.payload.text is None:
            return ''
        return self.payload.text

    async def to_recalled(self) -> Message:
        """
        Get the recalled message
        """
        if self.message_type() != MessageType.MESSAGE_TYPE_RECALLED:
            raise WechatyOperationError(
                'Can not call toRecalled() on message which is not'
                ' recalled type.')

        origin_message_id = self.text()
        if origin_message_id is None:
            raise WechatyPayloadError('Can not find recalled message')

        log.info('get recall message <%s>', origin_message_id)
        try:
            message = self.wechaty.Message.load(origin_message_id)
            await message.ready()
            return message
        except Exception as exception:
            error_info = 'can"t load or ready message payload {}'.format(
                str(exception.args)
            )

            log.error(error_info)
            raise WechatyOperationError(error_info)

    async def recall(self) -> bool:
        """
        Recall a message.
        """
        log.info('Message recall')
        success = await self.puppet.message_recall(self.message_id)
        return success

    @classmethod
    def load(cls, message_id: str) -> Message:
        """
        Create a Mobile Terminated Message
        """
        return cls(message_id)

    def type(self) -> MessageType:
        """
        Get the type from the message.
        :return:
        """
        return self.payload.type

    def is_self(self) -> bool:
        """
        Check if a message is sent by self
        :return:
        """
        user_id = self.wechaty.contact_id
        talker = self.talker()
        if talker is None:
            return False
        return talker.contact_id == user_id

    async def mention_list(self) -> List[Contact]:
        """
        Get message mentioned contactList.
        :return:
        """
        log.info('Message mention_list')
        room = self.room()
        if self.type() != MessageType.MESSAGE_TYPE_TEXT or room is None:
            return []

        # Use mention list if mention list is available
        # otherwise, process the message and get the mention list

        if self.payload is not None and self.payload.mention_ids is not None:
            async def id_to_contact(contact_id) -> Contact:
                contact = self.wechaty.Contact.load(contact_id)
                await contact.ready()
                return contact

            # TODO -> change to python async best practice
            contacts = [
                await id_to_contact(contact_id)
                for contact_id in self.payload.mention_ids]
            return contacts

        # TODO -> have to check that mention_id is not in room situation
        return []

    async def mention_text(self) -> str:
        """
        get mention text
        :return:
        """
        text = self.text()
        room = self.room()

        mention_list = await self.mention_list()

        if room is None or len(mention_list) <= 0:
            return text

        async def get_alias_or_name(member: Contact) -> str:
            if room is not None:
                alias = await room.alias(member)
                if alias is not None:
                    return alias
            return member.name

        # TODO -> change to python async best practice
        # flake8: disable=F841
        mention_names = [
            await get_alias_or_name(member)
            for member in mention_list]
        # TOD -> need to remove
        reversed(mention_names)

        # const textWithoutMention = mentionNameList.reduce((prev, cur) => {
        #     const escapedCur = escapeRegExp(cur)
        #     const regex = new RegExp(`@${escapedCur}(\u2005|\u0020|$)`)
        #     return prev.replace(regex, '')
        # }, text)

        # import re
        # from functools import reduce
        # def reg_replace(pre, cur) -> str:
        #     regex =
        return ''

    async def mention_self(self) -> bool:
        """
        Check if a message is mention self.
        :return:
        """
        self_id = self.wechaty.contact_id

        # check and ready for message payload
        await self.ready()

        # check by mention_ids not mention_list
        if self.payload is None or self.payload.mention_ids is None:
            return False
        return self_id in self.payload.mention_ids

    async def ready(self):
        """
        sync load message
        """
        log.debug('Message ready <%s>', self)
        if self.is_ready():
            return

        self.payload = await self.puppet.message_payload(self.message_id)

        if self.payload.from_id.strip() != '':
            talker = self.wechaty.Contact.load(self.payload.from_id)
            await talker.ready()
        if self.payload.room_id.strip() != '':
            room = self.wechaty.Room.load(self.payload.room_id)
            await room.ready()
        if self.payload.to_id.strip() != '':
            to_contact = self.wechaty.Contact.load(self.payload.to_id)
            await to_contact.ready()

    async def forward(self, to: Union[Room, Contact]):
        """
        doc
        :param to:
        :return:
        """
        log.info('forward() <%s>', to)
        if to is None:
            raise WechatyPayloadError('to param not found')
        try:
            if isinstance(to, Room):
                to_id = to.room_id
            elif isinstance(to, Contact):
                to_id = to.contact_id
            else:
                raise WechatyPayloadError(
                    'expected type is <Room, Contact>, but get <%s>'
                    % to.__class__)
            print(to_id)
            await self.puppet.message_forward(to_id, self.message_id)

        # pylint:disable=W0703
        except Exception as exception:
            message = 'Message forward error <%s>' % exception.args
            log.error(message)
            raise WechatyOperationError(message)

    def date(self) -> datetime:
        """
        Message sent date
        :return:
        """
        time = datetime.fromtimestamp(self.payload.time_stamp)
        return time

    def age(self) -> int:
        """
        Returns the message age in seconds.
        :return:
        """
        return (datetime.now() - self.date()).seconds // 1000

    async def to_file_box(self) -> FileBox:
        """
        Extract the Media File from the Message, and put it into the FileBox.

        File MessageType is : {
            MESSAGE_TYPE_ATTACHMENT,
            MESSAGE_TYPE_EMOTICON,
            MESSAGE_TYPE_IMAGE,
            MESSAGE_TYPE_VIDEO
        }
        """
        log.info('Message to FileBox')
        if self.type() not in SUPPORTED_MESSAGE_FILE_TYPES:
            raise WechatyOperationError(
                f'this type <{self.type().name}> message can"t be converted to '
                f'FileBox'
            )
        file_box = await self.puppet.message_file(self.message_id)
        return file_box

    def to_image(self) -> Image:
        """
        Extract the Image File from the Message, so that we can use
        different image sizes.
        :return:
        """
        log.info('Message to Image() for message %s', self.message_id)
        if self.type() != MessageType.MESSAGE_TYPE_IMAGE:
            raise WechatyOperationError(
                'current message type: %s, not image type'
                % self.type()
            )
        return self.wechaty.Image.create(self.message_id)

    async def to_contact(self) -> Contact:
        """
        Get Share Card of the Message
        Extract the Contact Card from the Message, and encapsulate it into
         Contact class
        :return:
        """
        log.info('Message to Contact')
        if self.type() != MessageType.MESSAGE_TYPE_CONTACT:
            raise WechatyOperationError(
                'current message type: %s, not contact type'
                % self.type()
            )

        contact_id = await self.puppet.message_contact(self.message_id)

        contact = self.wechaty.Contact.load(contact_id)
        await contact.ready()
        return contact

    async def to_url_link(self) -> UrlLink:
        """
        get url_link from message
        :return:
        """
        log.info('Message to UrlLink')
        if self.type() != MessageType.MESSAGE_TYPE_URL:
            raise WechatyOperationError(
                'current message type: %s, not url type'
                % self.type()
            )
        payload = await self.puppet.message_url(self.message_id)
        if payload is None:
            raise WechatyPayloadError(
                'can not get url_link_payload by message: %s'
                % self.message_id)
        return UrlLink(payload)

    async def to_mini_program(self) -> MiniProgram:
        """
        get message mini_program
        :return:
        """
        log.info('Message to MiniProgram <%s>', self.message_id)

        if self.type() != MessageType.MESSAGE_TYPE_MINI_PROGRAM:
            raise WechatyOperationError('not a mini_program type message')

        payload = await self.puppet.message_mini_program(
            self.message_id)
        if payload is None:
            raise WechatyPayloadError(
                'no miniProgram payload for message %s'
                % self.message_id
            )
        return MiniProgram(payload)



"""
Python Wechaty - https://github.com/wechaty/python-wechaty

Authors:    Alfred Huang (黃文超) <https://github.com/fish-ball>

2020-now @ Copyright Wechaty

Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an 'AS IS' BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class WechatyError(Exception):
    """ Wechaty error """

    def __init__(self, message, code=None, params=None):
        super().__init__(message, code, params)

        self.message = message
        self.code = code
        self.params = params

    def __str__(self):
        return repr(self)


class WechatyAccessoryBindingError(WechatyError):
    """ Raises when using Accessory classes in the wrong way """


class WechatyStatusError(WechatyError, AttributeError):
    """ Wechaty method calling o non-proper status (e.g. lack of await ready) """


class WechatyConfigurationError(WechatyError, AttributeError):
    """ Raises when configuration out of expected case """


class WechatyOperationError(WechatyError):
    """ Logical out of business error occurs when using wechaty """


class WechatyPluginError(WechatyError):
    """ Error occurs when using plugin """


class WechatyPayloadError(WechatyError, ValueError):
    """ Error occurs when the GRPC service return data out of expected """
